/* Copyright (C) 2012-2016 Free Software Foundation, Inc.

   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library.  If not, see
   <http://www.gnu.org/licenses/>.  */

/* Assumptions:
 *
 * ARMv8-a, AArch64
 * aligned accesses
 *
 */

#define dstin		x0
#define src		x1
#define count		x2
#define tmp1		x3
#define tmp1w		w3
#define tmp2		x4
#define tmp2w		w4
#define tmp3		x5
#define tmp3w		w5
#define dst		x6

#define	A_w		w7
#define	B_w		w9
#define	C_w		w11
#define	D_w		w13

#define A_l		x7
#define A_h		x8
#define B_l		x9
#define B_h		x10
#define C_l		x11
#define C_h		x12
#define D_l		x13
#define D_h		x14

#include <sysdep.h>

// FT2000 series
ENTRY_ALIGN (mesa_memcpy_align_128, 6)
	mov		dst, dstin

	cmp     count, #16
	b.lt    1f

	and     tmp3, dst, #0x0f
	and     tmp2, src, #0x0f
	cmp     tmp2, tmp3
	b.eq    L(DstEqSrc16)
1:
	orr     tmp1, dst, src
	tbnz    tmp1w, #0, L(DataAlign1)
	tbnz    tmp1w, #1, L(DataAlign2)
	tbnz    tmp1w, #2, L(DataAlign4)
	tbnz    tmp1w, #3, L(DataAlign8)
	b       L(DataAlign16)

L(DstEqSrc16):
	neg     tmp1, src
	ands    tmp1, tmp1, #15
	b.eq    L(DataAlign16)
	sub     count, count, tmp1

	tbz     tmp1, #0, 3f
	ldrb    A_w, [src], #1
	strb    A_w, [dst], #1
3:
	tbz     tmp1, #1, 2f
	ldrh    A_w, [src], #2
	strh    A_w, [dst], #2
2:
	tbz     tmp1, #2, 1f
	ldr     A_w, [src], #4
	str     A_w, [dst], #4
1:
	tbz     tmp1, #3, L(DataAlign16)
	ldr     A_l, [src], #8
	str     A_l, [dst], #8

L(DataAlign16):
3:
	cmp     count, #128
	b.ge    2f
	cmp     count, #64
	b.ge    1f
	b.lt    L(tail63)
2:
	ldp     A_l, A_h, [src], #16
	ldp     B_l, B_h, [src], #16
	ldp     C_l, C_h, [src], #16
	ldp     D_l, D_h, [src], #16

	stp     A_l, A_h, [dst], #16
	stp     B_l, B_h, [dst], #16
	stp     C_l, C_h, [dst], #16
	stp     D_l, D_h, [dst], #16
	sub     count, count, #64

1:
	ldp     A_l, A_h, [src], #16
	ldp     B_l, B_h, [src], #16
	ldp     C_l, C_h, [src], #16
	ldp     D_l, D_h, [src], #16

	stp     A_l, A_h, [dst], #16
	stp     B_l, B_h, [dst], #16
	stp     C_l, C_h, [dst], #16
	stp     D_l, D_h, [dst], #16

	sub     count, count, #64
	b       3b

L(tail63):
	ands    tmp1, count, #0x30
	b.eq    4f
	cmp     tmp1, #0x20
	b.eq    6f
	b.lt    5f
	ldp     A_l, A_h, [src], #16
	stp     A_l, A_h, [dst], #16
6:
	ldp     A_l, A_h, [src], #16
	stp     A_l, A_h, [dst], #16
5:
	ldp     A_l, A_h, [src], #16
	stp     A_l, A_h, [dst], #16
4:
	tbz     count, #3, 3f
	ldr     A_l, [src], #8
	str     A_l, [dst], #8
3:
	tbz     count, #2, 2f
	ldr     A_w, [src], #4
	str     A_w, [dst], #4
2:
	tbz     count, #1, 1f
	ldrh    A_w, [src], #2
	strh    A_w, [dst], #2
1:
	tbz     count, #0, 0f
	ldrb    A_w, [src], #1
	strb    A_w, [dst], #1
0:
	RET

END (mesa_memcpy_align_128)

// FT1500A series
ENTRY_ALIGN (mesa_memcpy_align_64, 6)
	mov		dst, dstin

	cmp     count, #8
	b.lt    1f

	and		tmp3, dst, #0x07
	and     tmp2, src, #0x07
	cmp     tmp2, tmp3
	b.eq    L(DstEqSrc8)
1:
	orr     tmp1, dst, src
	tbnz    tmp1w,  #0,   L(DataAlign1)
	tbnz    tmp1w,  #1,   L(DataAlign2)
	tbnz    tmp1w,  #2,   L(DataAlign4)
	b       L(DataAlign8)

L(DstEqSrc8):
	neg		tmp1, src
	ands    tmp1, tmp1, #7
	b.eq    L(DataAlign8)
	sub    	count, count, tmp1

	tbz		tmp1, #0, 2f
	ldrb    A_w, [src], #1
	strb    A_w, [dst], #1
2:
	tbz		tmp1, #1, 1f
	ldrh    A_w, [src], #2
	strh    A_w, [dst], #2
1:
	tbz		tmp1, #2, L(DataAlign8)
	ldr		A_w, [src], #4
	str		A_w, [dst], #4

L(DataAlign8):
3:
	cmp     count, #64
	b.ge    2f
	cmp     count, #32
	b.ge    1f
	b.lt    L(tail31)
2:
	ldr		A_l, [src], #8
	ldr		B_l, [src], #8
	ldr 	C_l, [src], #8
	ldr		D_l, [src], #8

	str		A_l, [dst], #8
	str		B_l, [dst], #8
	str		C_l, [dst], #8
	str		D_l, [dst], #8

	sub		count, count, #32

1:
	ldr		A_l, [src], #8
	ldr		B_l, [src], #8
	ldr		C_l, [src], #8
	ldr		D_l, [src], #8

	str		A_l, [dst], #8
	str		B_l, [dst], #8
	str		C_l, [dst], #8
	str		D_l, [dst], #8

	sub		count, count, #32
	b       3b

L(tail31):
	ands    tmp1, count, #0x18
	b.eq    3f
	cmp		tmp1, #0x10
	b.eq    5f
	b.lt    4f
	ldr		A_l, [src], #8
	str     A_l, [dst], #8
5:
	ldr		A_l, [src], #8
	str		A_l, [dst], #8
4:
	ldr		A_l, [src], #8
	str		A_l, [dst], #8
3:
	tbz		count, #2, 2f
	ldr		A_w, [src], #4
	str		A_w, [dst], #4
2:
	tbz		count, #1,1f
	ldrh    A_w, [src], #2
	strh    A_w, [dst], #2
1:
	tbz		count, #0, 0f
	ldrb    A_w, [src], #1
	strb    A_w, [dst], #1
0:
	RET

L(DataAlign4):
3:
	cmp     count, #32
	b.ge    2f
	cmp     count, #16
	b.ge    1f
	b.lt    L(tail15)
2:
	ldr    A_w, [src], #4
	ldr    B_w, [src], #4
	ldr    C_w, [src], #4
	ldr    D_w, [src], #4

	str    A_w, [dst], #4
	str    B_w, [dst], #4
	str    C_w, [dst], #4
	str    D_w, [dst], #4

	sub     count, count, #16
1:
	ldr    A_w, [src], #4
	ldr    B_w, [src], #4
	ldr    C_w, [src], #4
	ldr    D_w, [src], #4

	str    A_w, [dst], #4
	str    B_w, [dst], #4
	str    C_w, [dst], #4
	str    D_w, [dst], #4

	sub     count, count, #16
	b       3b

L(tail15):
	ands    tmp1, count, #0x0c
	b.eq    2f
	cmp		tmp1, #0x08
	b.eq    4f
	b.lt    3f
	ldr		A_w, [src], #4
	str		A_w, [dst], #4
4:
	ldr		A_w, [src], #4
	str		A_w, [dst], #4
3:
	ldr		A_w, [src], #4
	str		A_w, [dst], #4
2:
	tbz		count, #1, 1f
	ldrh    A_w, [src], #2
	strh    A_w, [dst], #2
1:
	tbz		count, #0, 0f
	ldrb    A_w, [src], #1
	strb    A_w, [dst], #1
0:
	RET

L(DataAlign2):
3:
	cmp     count, #16
	b.ge    2f
	cmp     count, #8
	b.ge    1f
	b.lt    L(tail7)
2:
	ldrh	A_w, [src], #2
	ldrh	B_w, [src], #2
	ldrh	C_w, [src], #2
	ldrh	D_w, [src], #2

	strh	A_w, [dst], #2
	strh	B_w, [dst], #2
	strh	C_w, [dst], #2
	strh	D_w, [dst], #2

	sub     count, count, #8
1:
	ldrh	A_w, [src], #2
	ldrh	B_w, [src], #2
	ldrh	C_w, [src], #2
	ldrh	D_w, [src], #2

	strh	A_w, [dst], #2
	strh	B_w, [dst], #2
	strh	C_w, [dst], #2
	strh	D_w, [dst], #2

	sub     count, count, #8
	b       3b

L(tail7):
	ands    tmp1, count, #0x06
	b.eq    1f
	cmp		tmp1, #0x04
	b.eq    3f
	b.lt    2f
	ldrh    A_w, [src], #2
	strh    A_w, [dst], #2
3:
	ldrh    A_w, [src], #2
	strh    A_w, [dst], #2
2:
	ldrh    A_w, [src], #2
	strh    A_w, [dst], #2
1:
	tbz	count, #0, 0f
	ldrb    A_w, [src], #1
	strb    A_w, [dst], #1
0:
	RET

L(DataAlign1):
3:
	cmp     count, #8
	b.ge    2f
	cmp     count, #4
	b.ge    1f
	b.lt    L(tail3)
2:
	ldrb    A_w, [src], #1
	ldrb    B_w, [src], #1
	ldrb    C_w, [src], #1
	ldrb    D_w, [src], #1

	strb    A_w, [dst], #1
	strb    B_w, [dst], #1
	strb    C_w, [dst], #1
	strb    D_w, [dst], #1

	sub     count, count, #4
1:
	ldrb    A_w, [src], #1
	ldrb    B_w, [src], #1
	ldrb    C_w, [src], #1
	ldrb    D_w, [src], #1

	strb    A_w, [dst], #1
	strb    B_w, [dst], #1
	strb    C_w, [dst], #1
	strb    D_w, [dst], #1

	sub     count, count, #4
	b       3b

L(tail3):
	tbz		count, #1, 1f
	ldrb    A_w, [src], #1
	ldrb    B_w, [src], #1
	strb    A_w, [dst], #1
	strb    B_w, [dst], #1
1:
	tbz		count, #0, 0f
	ldrb    A_w, [src], #1
	strb    A_w, [dst], #1
0:
	RET

END (mesa_memcpy_align_64)
