/* Copyright (C) 2012-2016 Free Software Foundation, Inc.

   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library.  If not, see
   <http://www.gnu.org/licenses/>.  */

/* Assumptions:
 *
 * ARMv8-a, AArch64
 * aligned accesses
 *
 */

#include <sysdep.h>

#define dstin		x0
#define val		w1
#define count		x2
#define tmp1		x3
#define tmp1w		w3
#define tmp2		x4
#define tmp2w		w4
#define tmp3		x5
#define tmp3w		w5
#define dst		x6

#define	A_w		w7
#define	B_w		w9
#define	C_w		w11
#define	D_w		w13

#define A_l		x7
#define A_h		x8
#define B_l		x9
#define B_h		x10
#define C_l		x11
#define C_h		x12
#define D_l		x13
#define D_h		x14

// FT2000 series
ENTRY_ALIGN (mesa_memset_align_128, 6)
	mov		dst, dstin

	dup		v0.16B, w1
	mov		A_l, v0.D[0]

	cmp     count, #16
	b.lt    L(tiny15)

	neg     tmp1, dst
	ands    tmp1, tmp1, #15
	b.eq    L(Align16)

	sub     count, count, tmp1
	tbz     tmp1, #0, 3f
	strb    A_w, [dst], #1
3:
	tbz     tmp1, #1, 2f
	strh    A_w, [dst], #2
2:
	tbz     tmp1, #2, 1f
	str     A_w, [dst], #4
1:
	tbz     tmp1, #3, L(Align16)
	str     A_l, [dst], #8

L(Align16):
3:
	cmp     count, #128
	b.ge    2f
	cmp     count, #64
	b.ge    1f
	b.lt    L(tail63)
2:
	stp     A_l, A_l, [dst], #16
	stp     A_l, A_l, [dst], #16
	stp     A_l, A_l, [dst], #16
	stp     A_l, A_l, [dst], #16

	sub     count, count, #64
1:
	stp     A_l, A_l, [dst], #16
	stp     A_l, A_l, [dst], #16
	stp     A_l, A_l, [dst], #16
	stp     A_l, A_l, [dst], #16

	sub     count, count, #64
	b       3b

L(tail63):
	ands    tmp2, count, #0x30
	b.eq    4f
	cmp     tmp2, #0x20
	b.eq    6f
	b.lt    5f
	stp     A_l, A_l, [dst], #16
6:
	stp     A_l, A_l, [dst], #16
5:
	stp     A_l, A_l, [dst], #16
4:
	tbz     count, #3, 3f
	str     A_l, [dst], #8
3:
	tbz     count, #2, 2f
	str     A_w, [dst], #4
2:
	tbz     count, #1, 1f
	strh    A_w, [dst], #2
1:
	tbz     count, #0, 0f
	strb    A_w, [dst], #1
0:
	RET

END (mesa_memset_align_128)

// FT1500A series
ENTRY_ALIGN (mesa_memset_align_64, 6)
	mov		dst, dstin

	dup		v0.16B, w1
	mov		A_l, v0.D[0]

	cmp		count, #8
	b.lt    L(tiny7)

	neg		tmp1, dst
	ands	tmp1, tmp1, #7
	b.eq    L(Align8)

	sub		count, count, tmp1
	tbz		tmp1, #0, 2f
	strb    A_w, [dst], #1
2:
	tbz		tmp1, #1, 1f
	strh    A_w, [dst], #2
1:
	tbz		tmp1, #2, L(Align8)
	str     A_w, [dst], #4

L(Align8):
3:
	cmp     count, #64
	b.ge    2f
	cmp     count, #32
	b.ge    1f
	b.lt    L(tail31)
2:
	str		A_l, [dst], #8
	str		A_l, [dst], #8
	str		A_l, [dst], #8
	str		A_l, [dst], #8

	sub		count, count, #32
1:
	str		A_l, [dst], #8
	str		A_l, [dst], #8
	str		A_l, [dst], #8
	str		A_l, [dst], #8

	sub		count, count, #32

	b       3b

L(tail31):
	ands    tmp2, count, #0x18
	b.eq    3f
	cmp     tmp2, #0x10
	b.eq    5f
	b.lt    4f
	str     A_l, [dst], #8
5:
	str     A_l, [dst], #8
4:
	str     A_l, [dst], #8
3:
	tbz     count, #2, 2f
	str     A_w, [dst], #4
2:
	tbz     count, #1, 1f
	strh    A_w, [dst], #2
1:
	tbz     count, #0, 0f
	strb    A_w, [dst], #1
0:
	RET

L(tiny15):
	tbz     count, #3, L(tiny7)
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1

L(tiny7):
	tbz     count, #2, 2f
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1
2:
	tbz     count, #1, 1f
	strb    A_w, [dst], #1
	strb    A_w, [dst], #1
1:
	tbz     count, #0, 0f
	strb    A_w, [dst], #1
0:
	RET

END (mesa_memset_align_64)
